//
// Created by alex on 11.06.24.
//

#ifndef CLIPRDR_FILE_H
#define CLIPRDR_FILE_H

#ifdef WITH_FUSE3
#define FUSE_USE_VERSION 30
#include <fuse3/fuse_lowlevel.h>
#elif WITH_FUSE2
#define FUSE_USE_VERSION 26
#include <fuse_lowlevel.h>
#endif             

#include <winpr/clipboard.h>
#include <freerdp/server/cliprdr.h>
#include <winpr/crypto.h>

typedef struct cliprdr_file_context CliprdrFileContext;
#if defined(WITH_FUSE2)||defined(WITH_FUSE3) 
typedef enum _FuseLowlevelOperationType
{
	FUSE_LL_OPERATION_NONE,
	FUSE_LL_OPERATION_LOOKUP,
	FUSE_LL_OPERATION_GETATTR,
	FUSE_LL_OPERATION_READ,
} FuseLowlevelOperationType;

typedef struct _CliprdrFuseFile CliprdrFuseFile;

struct _CliprdrFuseFile
{
	CliprdrFuseFile* parent;
	wArrayList* children;

	char* filename;
	char* filename_with_root;
	UINT32 list_idx;
	fuse_ino_t ino;

	BOOL is_directory;
	BOOL is_readonly;

	BOOL has_size;
	UINT64 size;

	BOOL has_last_write_time;
	UINT64 last_write_time_unix;

	BOOL has_clip_data_id;
	UINT32 clip_data_id;
};

typedef struct
{
	CliprdrFileContext* file_context;

	CliprdrFuseFile* clip_data_dir;

	BOOL has_clip_data_id;
	UINT32 clip_data_id;
} CliprdrFuseClipDataEntry;

typedef struct
{
	CliprdrFileContext* file_context;

	wArrayList* fuse_files;

	BOOL all_files;
	BOOL has_clip_data_id;
	UINT32 clip_data_id;
} FuseFileClearContext;

typedef struct
{
	FuseLowlevelOperationType operation_type;
	CliprdrFuseFile* fuse_file;
	fuse_req_t fuse_req;
	UINT32 stream_id;
} CliprdrFuseRequest;

typedef struct
{
	CliprdrFuseFile* parent;
	char* parent_path;
} CliprdrFuseFindParentContext;
#endif

typedef struct
{
	char* name;
	FILE* fp;
	INT64 size;
	CliprdrFileContext* context;
} CliprdrLocalFile;

typedef struct
{
	UINT32 lockId;
	BOOL locked;
	size_t count;
	CliprdrLocalFile* files;
	CliprdrFileContext* context;
} CliprdrLocalStream;

typedef BOOL (*pf_shadow_file_context_uninit)(CliprdrFileContext* file, CliprdrServerContext * cliprdr);
typedef BOOL (*pf_shadow_file_context_init)(CliprdrFileContext* file, CliprdrServerContext * cliprdr);
typedef UINT32 (*pf_shadow_file_context_remote_get_flags)(CliprdrFileContext* file);
typedef BOOL (*pf_shadow_file_context_has_local_support)(CliprdrFileContext* file);
typedef BOOL (*pf_shadow_file_context_remote_set_flags)(CliprdrFileContext* file, UINT32 flags);
typedef BOOL (*pf_shadow_file_context_set_locally_available)(CliprdrFileContext* file, BOOL available);
typedef UINT32 (*pf_shadow_file_context_current_flags)(CliprdrFileContext* file);
typedef BOOL (*pf_shadow_file_context_update_client_data)(CliprdrFileContext* file, const char* data, size_t size);
typedef BOOL (*pf_shadow_file_context_clear)(CliprdrFileContext* file);
typedef CliprdrFileContext* (*pf_shadow_file_context_new)(void* context);
typedef void (*pf_cshadow_file_context_free)(CliprdrFileContext* file);
typedef void* (*pf_shadow_file_context_get_context)(CliprdrFileContext* file);
typedef BOOL (*pf_shadow_file_context_update_server_data)(CliprdrFileContext* file_context, wClipboard* clip,const void* data, size_t size);
typedef UINT (*pf_shadow_file_context_notify_new_server_format_list)(CliprdrFileContext* file_context);
typedef UINT (*pf_shadow_file_context_notify_new_client_format_list)(CliprdrFileContext* file_context);

struct cliprdr_file_context
{
#if defined(WITH_FUSE2) || defined(WITH_FUSE3)  
	/* FUSE related**/
	HANDLE fuse_start_sync;
	HANDLE fuse_stop_sync;
	HANDLE fuse_thread;
	struct fuse_session* fuse_sess;

	#if FUSE_USE_VERSION < 30
	struct fuse_chan* ch;
	#endif     


	wHashTable* inode_table;
	wHashTable* clip_data_table;
	wHashTable* request_table;

	CliprdrFuseFile* root_dir;
	CliprdrFuseClipDataEntry* clip_data_entry_without_id;
	UINT32 current_clip_data_id;

	fuse_ino_t next_ino;
	UINT32 next_clip_data_id;
	UINT32 next_stream_id;
#endif

	/* File clipping */
	BOOL file_formats_registered;
	UINT32 file_capability_flags;

	UINT32 local_lock_id;

	wHashTable* local_streams;
	wLog* log;
	void* clipboard;
	CliprdrServerContext * context;
	char* path;
	char* exposed_path;
	BYTE server_data_hash[WINPR_SHA256_DIGEST_LENGTH];
	BYTE client_data_hash[WINPR_SHA256_DIGEST_LENGTH];
};
BOOL cliprdr_file_context_uninit(CliprdrFileContext* file, CliprdrServerContext * cliprdr);
BOOL cliprdr_file_context_init(CliprdrFileContext* file, CliprdrServerContext * cliprdr);
UINT32 cliprdr_file_context_remote_get_flags(CliprdrFileContext* file);
BOOL cliprdr_file_context_has_local_support(CliprdrFileContext* file);
BOOL cliprdr_file_context_remote_set_flags(CliprdrFileContext* file, UINT32 flags);
BOOL cliprdr_file_context_set_locally_available(CliprdrFileContext* file, BOOL available);
UINT32 cliprdr_file_context_current_flags(CliprdrFileContext* file);
BOOL cliprdr_file_context_update_client_data(CliprdrFileContext* file, const char* data, size_t size);
BOOL cliprdr_file_context_clear(CliprdrFileContext* file);
CliprdrFileContext* cliprdr_file_context_new(void* context);
void cliprdr_file_context_free(CliprdrFileContext* file);
void* cliprdr_file_context_get_context(CliprdrFileContext* file);
BOOL cliprdr_file_context_update_server_data(CliprdrFileContext* file_context, wClipboard* clip,const void* data, size_t size);
UINT cliprdr_file_context_notify_new_server_format_list(CliprdrFileContext* file_context);
UINT cliprdr_file_context_notify_new_client_format_list(CliprdrFileContext* file_context);



#endif //CLIPRDR_FILE_H
